from sparta.memcached.base import SpartaCache


class SpartaCacheDummy(SpartaCache):
    """
    Dummy implementation, useful to disable caching without breaking code deps.
    """

    def add(self, key, *args, **kwargs):
        pass

    def append(self, key, *args, **kwargs):
        pass

    def cas(self, key, *args, **kwargs):
        pass

    def decr(self, key, *args, **kwargs):
        pass

    def delete(self, key, *args, **kwargs):
        pass

    def get(self, key, *args, **kwargs):
        pass

    def gets(self, key, *args, **kwargs):
        pass

    def incr(self, key, *args, **kwargs):
        pass

    def prepend(self, key, *args, **kwargs):
        pass

    def replace(self, key, *args, **kwargs):
        pass

    def set(self, key, *args, **kwargs):
        pass

    def touch(self, key, *args, **kwargs):
        pass



import abc
import functools
import inspect


class SpartaCache(abc.ABC):
    def __init__(
        self,
        key_prefix=None,
    ):
        self.key_prefix = key_prefix if isinstance(key_prefix, str) else str(key_prefix)

    def _map_key(self, key):
        return self.key_prefix + key if self.key_prefix else key

    @abc.abstractmethod
    def add(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def append(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def cas(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def decr(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def delete(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def get(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def gets(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def incr(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def prepend(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def replace(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def set(self, key, *args, **kwargs):
        pass

    @abc.abstractmethod
    def touch(self, key, *args, **kwargs):
        pass

    def func_decorator(
        self,
        key: str,
        time: int = None,
        expiration: int = None,  # TODO remove expiration on next release
    ):
        """
        Use to annotate any method to cache its result.
        @param key: the cache key
        @param time: the cache expiration for current key
        @param expiration: DEPRECATED: use `time` instead

        Example:
            ```
            @cache.func_decorator("my-key", time=3600):
            def do_stuff():
                ...
            ```
        The code here above will store in cache (with expiration 1h) the result of `do_stuff`. Any following
        invocation of `do_stuff`, withing the cache expiration time, will return the cached value.
        """

        def from_cache():
            return self.get(key)

        def to_cache(value):
            self.set(key, value, time if time is not None else expiration)

        def decorator(f):

            if inspect.iscoroutinefunction(f):

                async def async_wrapped_function(*args, **kwargs):
                    _result = from_cache()
                    if _result is None:
                        _result = await f(*args, **kwargs)
                        to_cache(_result)
                    return _result

                return functools.update_wrapper(async_wrapped_function, f)

            else:

                def wrapped_function(*args, **kwargs):
                    _result = from_cache()
                    if _result is None:
                        _result = f(*args, **kwargs)
                        to_cache(_result)
                    return _result

                return functools.update_wrapper(wrapped_function, f)

        return decorator

    def atomic_append(self, key, value, _retries=3, *args, **kwargs):
        """
        Required cas behavior!
        Assures atomic append to a list of objects (only tuple is supported, not list).
        Solution copied from https://stackoverflow.com/a/27468294/20230162
        """
        try:
            _values, cas_id = self.gets(key)
        except ValueError as e:
            raise RuntimeError("atomic_append requires cas behavior") from e
        if _values is None:
            self.set(key, (value,), *args, **kwargs)
        else:
            if value not in _values:
                ok = self.cas(key, _values + (value,), cas_id, *args, **kwargs)
                if not ok and _retries > 1:
                    self.atomic_append(key, value, _retries - 1, *args, **kwargs)



import pylibmc

from sparta.memcached.base import SpartaCache


class SpartaCacheClient(pylibmc.Client, SpartaCache):
    """
    Extends pylibmc.Client to support key_prefix.
    Also see https://github.com/memcached/memcached/wiki/Commands.
    """

    def __init__(
        self,
        servers,
        behaviors=None,
        binary=False,
        username=None,
        password=None,
        key_prefix=None,
    ):
        super().__init__(servers, behaviors, binary, username, password)
        SpartaCache.__init__(self, key_prefix)

    def add(self, key, *args, **kwargs):
        return super().add(self._map_key(key), *args, **kwargs)

    def add_multi(self, *args, **kwargs):
        raise NotImplementedError("multi operations not supported")

    def append(self, key, *args, **kwargs):
        return super().append(self._map_key(key), *args, **kwargs)

    def cas(self, key, *args, **kwargs):
        return super().cas(self._map_key(key), *args, **kwargs)

    def decr(self, key, *args, **kwargs):
        return super().decr(self._map_key(key), *args, **kwargs)

    def delete(self, key, *args, **kwargs):
        return super().delete(self._map_key(key), *args, **kwargs)

    def delete_multi(self, *args, **kwargs):
        raise NotImplementedError("multi operations not supported")

    def get(self, key, *args, **kwargs):
        return super().get(self._map_key(key), *args, **kwargs)

    def gets(self, key, *args, **kwargs):
        return super().gets(self._map_key(key), *args, **kwargs)

    def get_multi(self, *args, **kwargs):
        raise NotImplementedError("multi operations not supported")

    def incr(self, key, *args, **kwargs):
        return super().incr(self._map_key(key), *args, **kwargs)

    def incr_multi(self, *args, **kwargs):
        raise NotImplementedError("multi operations not supported")

    def prepend(self, key, *args, **kwargs):
        return super().prepend(self._map_key(key), *args, **kwargs)

    def replace(self, key, *args, **kwargs):
        return super().replace(self._map_key(key), *args, **kwargs)

    def set(self, key, *args, **kwargs):
        return super().set(self._map_key(key), *args, **kwargs)

    def set_multi(self, *args, **kwargs):
        raise NotImplementedError("multi operations not supported")

    def touch(self, key, *args, **kwargs):
        return super().touch(self._map_key(key), *args, **kwargs)



import ipaddress
import socket
import typing
from urllib.parse import urlsplit


# Solution copied from this lab https://www.cloudskillsboost.google/focuses/615?parent=catalog

def _is_valid_ip(addr, _class: typing.Any = ipaddress.IPv4Network):
    try:
        _class(addr)
        return True
    except ValueError:
        return False


def is_valid_ipv4(addr):
    return _is_valid_ip(addr)


def is_valid_ipv6(addr):
    return _is_valid_ip(addr, _class=ipaddress.IPv6Network)


def _append_port(addr: str, port: typing.Optional[int]):
    return f"{addr}:{port}" if port else addr


def _resolve_dns(server: str) -> typing.List[str]:
    try:
        # urlparse() and urlsplit() insists on absolute URLs starting with "//"
        result = urlsplit('//' + server)
        addr = result.hostname
        port = result.port
        if is_valid_ipv4(addr):  # memcached supports only IPv4
            return [_append_port(addr, port)]
        _, _, ips = socket.gethostbyname_ex(addr)
        return [_append_port(ip, port) for ip in ips]
    except Exception as error:
        raise ValueError(f"Invalid server '{server}'") from error


def resolve_dns(
        servers: typing.Union[str, typing.List[str]]
) -> typing.List[str]:
    if isinstance(servers, str):
        return _resolve_dns(servers)
    return [x for server in servers for x in _resolve_dns(server)]



import pylibmc

from sparta.memcached.base import SpartaCache


class SpartaCacheWrapper(SpartaCache):
    """
    Wrapper around pylibmc.Client to support key_prefix.
    Also see https://github.com/memcached/memcached/wiki/Commands.
    """

    def __init__(
        self,
        client: pylibmc.Client,
        key_prefix: str,
    ):
        super().__init__(key_prefix)
        self.client = client

    def __repr__(self):
        return "<%s for %s>" % (self.__class__.__name__, self.client)

    def __str__(self):
        return "<%s for %s>" % (self.__class__.__name__, self.client)

    def __getitem__(self, key):
        return self.client.__getitem__(self._map_key(key))

    def __setitem__(self, key, value):
        return self.client.__setitem__(self._map_key(key), value)

    def __delitem__(self, key):
        return self.client.__delitem__(self._map_key(key))

    def __contains__(self, key):
        return self.client.__contains__(self._map_key(key))

    def add(self, key, *args, **kwargs):
        return self.client.add(self._map_key(key), *args, **kwargs)

    def append(self, key, *args, **kwargs):
        return self.client.append(self._map_key(key), *args, **kwargs)

    def cas(self, key, *args, **kwargs):
        return self.client.cas(self._map_key(key), *args, **kwargs)

    def decr(self, key, *args, **kwargs):
        return self.client.decr(self._map_key(key), *args, **kwargs)

    def delete(self, key, *args, **kwargs):
        return self.client.delete(self._map_key(key), *args, **kwargs)

    def get(self, key, *args, **kwargs):
        return self.client.get(self._map_key(key), *args, **kwargs)

    def gets(self, key, *args, **kwargs):
        return self.client.gets(self._map_key(key), *args, **kwargs)

    def incr(self, key, *args, **kwargs):
        return self.client.incr(self._map_key(key), *args, **kwargs)

    def prepend(self, key, *args, **kwargs):
        return self.client.prepend(self._map_key(key), *args, **kwargs)

    def replace(self, key, *args, **kwargs):
        return self.client.replace(self._map_key(key), *args, **kwargs)

    def set(self, key, *args, **kwargs):
        return self.client.set(self._map_key(key), *args, **kwargs)

    def touch(self, key, *args, **kwargs):
        return self.client.touch(self._map_key(key), *args, **kwargs)
